# -*- coding: utf-8 -*-
"""
Routines for parsing LEMI field spreadsheets(i.e. install and demobilization
sheets) and validating metadata;
Templates for the spreadsheets are provided to the PIs and their formatting is
hard coded.
The metadata for a complete  MT dataset was structured into the following
categories: Survey, Station, Run, Electric, Magnetic, and Auxiliary

Maeva Pourpoint - IRIS/PASSCAL
"""
from __future__ import annotations

import logging
import logging.config
import re

from configobj import ConfigObj
from dataclasses import dataclass, field
from obspy import UTCDateTime
from pathlib import Path
from typing import Optional, Set

from lemi2seed.lemi_data import (SAMPLING_RATE, VALID_COMPONENTS,
                                 CHANNEL_NAMING_CONVENTION)
from lemi2seed.utils import (check_email_formatting, check_instrument_specs,
                             check_serial_number, is_valid_uri, str2list,
                             ELECTRODE_DEFAULT, MSG_E_CHA)

# Read lemi2seed config file
config_file = Path(__file__).parent.joinpath('config.ini')
config = ConfigObj(str(config_file))

# Read logging config file
log_file_path = Path(__file__).parent.joinpath('logging.conf')
logging.config.fileConfig(log_file_path)
# Create logger
logger = logging.getLogger(__name__)

E_COMPONENT_NAMING_CONVENTION = {'North-South': 'Ex', 'South-North': 'Ex',
                                 'East-West': 'Ey', 'West-East': 'Ey'}


@dataclass
class BaseMetadata:
    """
    This class handles the metadata attributes shared at all levels: survey;
    station; run; electric; magnetic; auxiliary.
    """
    # List of metadata fields that are required for archiving but "missing"
    # because not provided by the user.
    metadata_missing: Set[str] = field(default_factory=set)
    # List of metadata fields that are invalid
    metadata_invalid: Set[str] = field(default_factory=set)


@dataclass
class BaseSurvey(BaseMetadata):
    """
    This class handles the metadata attributes shared at the survey and station
    levels.
    """
    restricted_status: Optional[str] = field(default=None,
                                             metadata={'xml_id': ['restricted_status'],
                                                       'required': True,
                                                       'gui': True})
    time_period_end: Optional[UTCDateTime] = field(default=None,
                                                   metadata={'xml_id': ['end_date'],
                                                             'required': True,
                                                             'gui': True})
    time_period_start: Optional[UTCDateTime] = field(default=None,
                                                     metadata={'xml_id': ['start_date'],
                                                               'required': True,
                                                               'gui': True})

    def validate_time_period_end(self, metadata_input: UTCDateTime, data_input: UTCDateTime) -> bool:
        if isinstance(metadata_input, UTCDateTime) and metadata_input >= data_input:
            return True
        else:
            logger.error("End date of {} should be in UTC and greater than "
                         "the data acquisition end time."
                         .format(self.__class__.__name__.split('_')[0]))
            return False

    def validate_time_period_start(self, metadata_input: UTCDateTime, data_input: UTCDateTime) -> bool:
        if isinstance(metadata_input, UTCDateTime) and metadata_input <= data_input:
            return True
        else:
            logger.error("Start date of {} should be in UTC and lower than "
                         "the data acquisition start time."
                         .format(self.__class__.__name__.split('_')[0]))
            return False


@dataclass
class Survey(BaseSurvey):
    """This class handles all the metadata attributes at the survey level."""
    acquired_by_author: Optional[str] = field(default=None,
                                              metadata={'xml_id': ['comments', 'mt.survey.acquired_by.author'],
                                                        'required': False,
                                                        'gui': True})
    acquired_by_comments: Optional[str] = field(default=None,
                                                metadata={'xml_id': ['comments', 'mt.survey.acquired_by.comments'],
                                                          'required': False,
                                                          'gui': True})
    archive_network: Optional[str] = field(default=None,
                                           metadata={'required': True,
                                                     'gui': True})
    citation_dataset_doi: Optional[str] = field(default=None,
                                                metadata={'xml_id': ['identifiers'],
                                                          'required': True,
                                                          'gui': True})
    comments: Optional[str] = field(default=None,
                                    metadata={'xml_id': ['comments', 'mt.survey.comments'],
                                              'required': False,
                                              'gui': True})
    country: Optional[str] = field(default=None,
                                   metadata={'xml_id': ['comments', 'mt.survey.country'],
                                             'required': False,
                                             'gui': True})
    geographic_name: Optional[str] = field(default=None,
                                           metadata={'xml_id': ['comments', 'mt.survey.geographic_name'],
                                                     'required': True,
                                                     'gui': True})
    name: Optional[str] = field(default=None,
                                metadata={'xml_id': ['description'],
                                          'required': True,
                                          'gui': True})
    project: Optional[str] = field(default=None,
                                   metadata={'xml_id': ['comments', 'mt.survey.project'],
                                             'required': True,
                                             'gui': True})
    project_lead_author: Optional[str] = field(default=None,
                                               metadata={'xml_id': ['operators.contacts.names'],
                                                         'required': False,
                                                         'gui': True})
    project_lead_email: Optional[str] = field(default=None,
                                              metadata={'xml_id': ['operators.contacts.emails'],
                                                        'required': False,
                                                        'gui': True})
    project_lead_organization: Optional[str] = field(default=None,
                                                     metadata={'xml_id': ['operators.agency'],
                                                               'required': False,
                                                               'gui': True})
    summary: Optional[str] = field(default=None,
                                   metadata={'xml_id': ['comments', 'mt.survey.summary'],
                                             'required': False,
                                             'gui': True})

    @staticmethod
    def validate_archive_network(metadata_input: str) -> bool:
        try:
            valid = re.match(r"^\w{2}$", metadata_input)
        except TypeError:
            logger.error("The network code should be a string.")
            return False
        else:
            if valid is None:
                logger.error("The network code should be two alphanumeric "
                             "character long.")
            return bool(valid)

    @staticmethod
    def validate_citation_dataset_doi(metadata_input: str) -> bool:
        # Proper formatting for DOI?
        try:
            dois = str2list(metadata_input)
        except AttributeError:
            logger.error("The DOI number(s) should be a string.")
            return False
        else:
            if not all([is_valid_uri(doi) for doi in dois]):
                logger.error("Invalid DOI(s). The DOI number(s) provided by "
                             "the archive should be strings formatted as "
                             "follows: 'scheme: path'.")
                return False
            return True

    @staticmethod
    def validate_project_lead_email(metadata_input: str) -> bool:
        if metadata_input is not None:
            try:
                emails = str2list(metadata_input)
            except AttributeError:
                logger.error("The project lead email(s) should be a string.")
                return False
            return all([check_email_formatting(email) for email in emails])
        return True


@dataclass
class BaseStation(BaseMetadata):
    """
    This class handles the metadata attributes shared at the station and
    channel levels.
    """
    data_type: str = field(default='LPMT',
                           metadata={'xml_id': ['comments', 'mt.station.data_type'],
                                     'required': True,
                                     'gui': False})
    elevation: Optional[float] = field(default=None,
                                       metadata={'required': True,
                                                 'gui': True})
    latitude: Optional[float] = field(default=None,
                                      metadata={'required': True,
                                                'gui': True})
    longitude: Optional[float] = field(default=None,
                                       metadata={'required': True,
                                                 'gui': True})

    @staticmethod
    def validate_geographics(param: tuple, metadata_input: float) -> bool:
        geo, min_range, max_range, min_val, max_val, units, msg = param
        try:
            metadata_input = float(metadata_input)
        except (TypeError, ValueError):
            logger.error("{} should be a float. {}".format(geo, msg))
            return False
        else:
            if metadata_input < min_range or metadata_input > max_range:
                logger.error("Unexpected {0}! The {0} should be between {1}"
                             "{3} and {2}{3}.".format(geo, min_range,
                                                      max_range, units))
                return False
            elif abs(metadata_input) < min_val or abs(metadata_input) > max_val:
                logger.error("Unexpected {0}! Provided {0} should "
                             "roughly match the {0} recorded by the on-site "
                             "GPS.".format(geo))
                return False
            return True

    def validate_elevation(self, metadata_input: float, data_input: float) -> bool:
        min_elev, max_elev = [abs(data_input) * x for x in [0.99, 1.01]]  # reasonable?
        param = ("Station elevation", -500, 8500, min_elev, max_elev, 'm', '')
        return BaseStation.validate_geographics(param, metadata_input)

    def validate_latitude(self, metadata_input: float, data_input: float) -> bool:
        min_lat, max_lat = [abs(data_input) * x for x in [0.995, 1.005]]  # reasonable?
        param = ("Station latitude", -90, 90, min_lat, max_lat, '°', '')
        return BaseStation.validate_geographics(param, metadata_input)

    def validate_longitude(self, metadata_input: float, data_input: float) -> bool:
        min_lon, max_lon = [abs(data_input) * x for x in [0.995, 1.005]]  # reasonable?
        param = ("Station longitude", -180, 180, min_lon, max_lon, '°', '')
        return BaseStation.validate_geographics(param, metadata_input)


@dataclass
class Station_(BaseSurvey, BaseStation):
    """This class handles all the metadata attributes at the station level."""
    archive_id: Optional[str] = field(default=None,
                                      metadata={'required': True,
                                                'gui': True})
    comments: Optional[str] = field(default=None,
                                    metadata={'xml_id': ['comments', 'mt.station.comments'],
                                              'required': False,
                                              'gui': True})
    declination_model: Optional[str] = field(default=None,
                                             metadata={'xml_id': ['comments', 'mt.station.location.declination.model'],
                                                       'required': True,
                                                       'gui': True})
    declination_value: Optional[float] = field(default=None,
                                               metadata={'xml_id': ['comments', 'mt.station.location.declination.value'],  # noqa: E501
                                                         'required': True,
                                                         'gui': True})
    geographic_name: Optional[str] = field(default=None,
                                           metadata={'xml_id': ['site'],
                                                     'required': False,
                                                     'gui': True})
    orientation_method: Optional[str] = field(default=None,
                                              metadata={'xml_id': ['comments', 'mt.station.orientation.method'],
                                                        'required': False,
                                                        'gui': True})
    orientation_reference_frame: Optional[str] = field(default=None,
                                                       metadata={'xml_id': ['comments', 'mt.station.orientation.reference_frame'],  # noqa: E501
                                                                 'required': True,
                                                                 'gui': True})
    run_list: str = field(default='a',
                          metadata={'xml_id': ['comments', 'mt.station.run_list'],
                                    'required': True,
                                    'gui': False})
    software_author: str = field(default=config['software']['author'],
                                 metadata={'xml_id': ['comments', 'mt.station.provenance.software.author'],
                                           'required': False,
                                           'gui': False})
    software_name: str = field(default=config['software']['name'],
                               metadata={'xml_id': ['comments', 'mt.station.provenance.software.name'],
                                         'required': False,
                                         'gui': False})
    software_version: str = field(default=config['software']['version'],
                                  metadata={'xml_id': ['comments', 'mt.station.provenance.software.version'],
                                            'required': False,
                                            'gui': False})
    submitter_author: Optional[str] = field(default=None,
                                            metadata={'xml_id': ['comments', 'mt.station.provenance.submitter.author'],
                                                      'required': False,
                                                      'gui': True})
    submitter_email: Optional[str] = field(default=None,
                                           metadata={'xml_id': ['comments', 'mt.station.provenance.submitter.email'],
                                                     'required': False,
                                                     'gui': True})
    submitter_organization: Optional[str] = field(default=None,
                                                  metadata={'xml_id': ['comments', 'mt.station.provenance.submitter.organization'],  # noqa: E501
                                                            'required': False,
                                                            'gui': True})

    @staticmethod
    def validate_archive_id(metadata_input: str) -> bool:
        try:
            valid = re.match(r"^\w{3,5}$", metadata_input)
        except TypeError:
            logger.error("The station name should be a string.")
            return False
        else:
            if valid is None:
                logger.error("The station name should be between three and "
                             "five alphanumeric character long.")
            return bool(valid)

    def validate_declination_value(self, metadata_input: float) -> bool:
        param = ("Declination angle", -90, 90, -99999, 99999, '°', None)
        return BaseStation.validate_geographics(param, metadata_input)

    @staticmethod
    def validate_submitter_email(metadata_input: str) -> bool:
        if metadata_input is not None:
            return check_email_formatting(metadata_input)
        return True


@dataclass
class Run(BaseMetadata):
    """This class handles all the metadata attributes at the run level."""
    acquired_by_author: Optional[str] = field(default=None,
                                              metadata={'xml_id': ['comments', 'mt.run:a.acquired_by.author'],
                                                        'required': False,
                                                        'gui': True})
    acquired_by_comments: Optional[str] = field(default=None,
                                                metadata={'xml_id': ['comments', 'mt.run:a.acquired_by.comments'],
                                                          'required': False,
                                                          'gui': True})
    components_recorded: Optional[str] = field(default=None,
                                               metadata={'required': True,
                                                         'gui': True})
    comments: Optional[str] = field(default=None,
                                    metadata={'xml_id': ['comments', 'mt.run:a.comments'],
                                              'required': False,
                                              'gui': True})
    data_type: str = field(default='LPMT',
                           metadata={'xml_id': ['type'],
                                     'required': True,
                                     'gui': False})
    datalogger_type: str = field(default='type: long-period 32-bit',
                                 metadata={'xml_id': ['description'],
                                           'required': False,
                                           'gui': False})
    datalogger_manufacturer: str = field(default='LEMI LLC.',
                                         metadata={'xml_id': ['manufacturer'],
                                                   'required': True,
                                                   'gui': False})
    datalogger_model: str = field(default='LEMI-424',
                                  metadata={'xml_id': ['model'],
                                            'required': True,
                                            'gui': True})
    datalogger_serial_number: Optional[str] = field(default=None,
                                                    metadata={'xml_id': ['serial_number'],
                                                              'required': True,
                                                              'gui': True})
    metadata_by_author: Optional[str] = field(default=None,
                                              metadata={'xml_id': ['comments', 'mt.run:a.metadata_by.author'],
                                                        'required': False,
                                                        'gui': True})
    metadata_by_comments: Optional[str] = field(default=None,
                                                metadata={'xml_id': ['comments', 'mt.run:a.metadata_by.comments'],
                                                          'required': False,
                                                          'gui': True})
    resource_id: Optional[str] = field(default=None,
                                       metadata={'xml_id': ['resource_id'],
                                                 'required': True,
                                                 'gui': False})
    time_period_end: Optional[UTCDateTime] = field(default=None,
                                                   metadata={'xml_id': ['removal_date'],
                                                             'required': True,
                                                             'gui': True})
    time_period_start: Optional[UTCDateTime] = field(default=None,
                                                     metadata={'xml_id': ['installation_date'],
                                                               'required': True,
                                                               'gui': True})

    @property
    def run_id(self) -> str:
        return self.resource_id.split(':')[1]  # type: ignore

    def validate_components_recorded(self, metadata_input: str) -> bool:
        """
        The following channels: Te, Tf, Ui, Sn, Fq, Ce (Sn: Satellite number,
        Fq: GPS Fix quality, Ce: Clock error) are recorded by default.
        """
        try:
            components = str2list(metadata_input)
        except AttributeError:
            logger.error("The list of components recorded for run '{}' "
                         "should be a string.".format(self.run_id))
            return False
        else:
            if not all([x in VALID_COMPONENTS for x in components]):
                logger.error("Some of the listed components for run '{}' are "
                             "not valid. List of valid components: {}."
                             .format(self.run_id, VALID_COMPONENTS))
                return False
            return True

    def validate_datalogger_serial_number(self, metadata_input: str) -> bool:
        equipment = "data logger (run {})".format(self.run_id)
        return check_serial_number(metadata_input, equipment)

    def validate_time_period_end(self, metadata_input: UTCDateTime, data_input: UTCDateTime) -> bool:
        if not (isinstance(metadata_input, UTCDateTime) and metadata_input <= data_input):
            logger.error("End date of run {} should be in UTC and lower than "
                         "the data acquisition end time."
                         .format(self.run_id))
            return False
        return True

    def validate_time_period_start(self, metadata_input: UTCDateTime, data_input: UTCDateTime) -> bool:
        if not (isinstance(metadata_input, UTCDateTime) and metadata_input >= data_input):
            logger.error("Start date of run {} should be in UTC and greater "
                         "than the data acquisition start time."
                         .format(self.run_id))
            return False
        return True


@dataclass
class BaseChannel(BaseStation):
    """
    This class handles the metadata attributes shared at the electric, magnetic
    and auxiliary levels.
    """
    channel_name: Optional[str] = field(default=None,
                                        metadata={'required': True,
                                                  'gui': False})
    component: Optional[str] = field(default=None,
                                     metadata={'xml_id': ['alternate_code'],
                                               'required': True,
                                               'gui': False})
    data_type: str = field(default='GEOPHYSICAL',
                           metadata={'xml_id': ['types'],
                                     'required': True,
                                     'gui': False})
    depth: int = field(default=0,
                       metadata={'required': True,
                                 'gui': False})
    instrument_manufacturer: Optional[str] = field(default=None,
                                                   metadata={'required': True,
                                                             'gui': True})
    instrument_model: Optional[str] = field(default=None,
                                            metadata={'required': True,
                                                      'gui': True})
    instrument_specs: Optional[str] = field(default=None,
                                            metadata={'required': True,
                                                      'gui': True})
    instrument_type: Optional[str] = field(default=None,
                                           metadata={'required': False,
                                                     'gui': True})
    location_code: str = field(default='',
                               metadata={'required': True,
                                         'gui': False})
    measurement_azimuth: Optional[float] = field(default=None,
                                                 metadata={'xml_id': ['azimuth'],
                                                           'required': True,
                                                           'gui': True})
    measurement_tilt: Optional[float] = field(default=None,
                                              metadata={'xml_id': ['dip'],
                                                        'required': True,
                                                        'gui': True})
    run_id: Optional[str] = field(default=None,
                                  metadata={'required': True,
                                            'gui': False})
    sample_rate: float = field(default=SAMPLING_RATE,
                               metadata={'xml_id': ['sample_rate'],
                                         'required': True,
                                         'gui': True})

    def validate_measurement_azimuth(self, metadata_input: float) -> bool:
        geo, msg = ["Azimuth angle", '']
        if isinstance(self, Electric):
            geo = ("Azimuth angle of electrode pair (run '{}' and channel "
                   "number '{}')".format(self.run_id, self.channel_number))  # type: ignore
            msg = MSG_E_CHA
        param = (geo, 0, 360, -99999, 99999, '°', msg)
        return BaseStation.validate_geographics(param, metadata_input)

    def validate_measurement_tilt(self, metadata_input: float) -> bool:
        geo, msg = ["Tilt angle", '']
        if isinstance(self, Electric):
            geo = ("Tilt angle of electrode pair (run '{}' and channel '{}')"
                   .format(self.run_id, self.channel_number))  # type: ignore
            msg = MSG_E_CHA
        param = (geo, 0, 180, -99999, 99999, '°', msg)
        return BaseStation.validate_geographics(param, metadata_input)


@dataclass
class Electric(BaseChannel):
    """This class handles all the metadata attributes at the electric level."""
    channel_number: Optional[str] = field(default=None,
                                          metadata={'xml_id': ['comments', 'mt.electric.channel_number'],
                                                    'required': True,
                                                    'gui': True})
    contact_resistance_end: Optional[float] = field(default=None,
                                                    metadata={'xml_id': ['comments', 'mt.electric.contact_resistance.end'],  # noqa: E501
                                                              'required': False,
                                                              'gui': True})
    contact_resistance_start: Optional[float] = field(default=None,
                                                      metadata={'xml_id': ['comments', 'mt.electric.contact_resistance.start'],  # noqa: E501
                                                                'required': False,
                                                                'gui': True})
    dc_end: Optional[float] = field(default=None,
                                    metadata={'xml_id': ['comments', 'mt.electric.dc.end'],
                                              'required': False,
                                              'gui': True})
    dc_start: Optional[float] = field(default=None,
                                      metadata={'xml_id': ['comments', 'mt.electric.dc.start'],
                                                'required': False,
                                                'gui': True})
    dipole_length: Optional[float] = field(default=None,
                                           metadata={'required': True,
                                                     'gui': True})
    negative_electrode_direction: Optional[str] = field(default=None,
                                                        metadata={'required': True,
                                                                  'gui': True})
    negative_electrode_serial_number: Optional[str] = field(default=None,
                                                            metadata={'required': True,
                                                                      'gui': True})
    positive_electrode_direction: Optional[str] = field(default=None,
                                                        metadata={'required': True,
                                                                  'gui': True})
    positive_electrode_serial_number: Optional[str] = field(default=None,
                                                            metadata={'required': True,
                                                                      'gui': True})

    def set_electric_component_channel(self) -> None:
        try:
            direction = "-".join([self.positive_electrode_direction, self.negative_electrode_direction])  # type: ignore
        except TypeError:
            logger.error("The direction of the positive and negative electrodes "
                         "should be either: North, South, East or West. Please, "
                         "provide electrode direction(s) for run '{}' and "
                         "channel number '{}'! {}"
                         .format(self.run_id, self.channel_number, MSG_E_CHA))
            self.metadata_invalid.update({'positive_electrode_direction',
                                          'negative_electrode_direction'})
        else:
            component = E_COMPONENT_NAMING_CONVENTION.get(direction)
            if component is None:
                if self.positive_electrode_direction == self.negative_electrode_direction:
                    logger.error("The direction of the positive and negative "
                                 "electrodes in a given pair cannot be the "
                                 "same (check run '{}' and channel number "
                                 "'{}')!".format(self.run_id, self.channel_number))
                logger.warning("By convention, one electrode pair is "
                               "installed in a north-south direction and the "
                               "other pair in a  east-west direction (check "
                               "run '{}' and channel number '{}')!"
                               .format(self.run_id, self.channel_number))
                self.metadata_invalid.update({'positive_electrode_direction',
                                              'negative_electrode_direction'})
            else:
                self.component = component
                self.channel_name = CHANNEL_NAMING_CONVENTION[component]
                self.metadata_invalid.discard('positive_electrode_direction')
                self.metadata_invalid.discard('negative_electrode_direction')

    def set_electrode_info(self) -> None:
        if self.instrument_specs == ELECTRODE_DEFAULT:
            self.instrument_manufacturer = "Borin"
            self.instrument_model = "STELTH 4"
            self.instrument_type = "Silver-Silver Chloride"
        elif self.instrument_specs is not None:
            specs = [x.split(':')[-1].strip() for x in self.instrument_specs.split('-')]
            self.instrument_manufacturer, self.instrument_model, self.instrument_type = specs

    def validate_instrument_specs(self, metadata_input: str) -> bool:
        equipment = 'electrode'
        channel_info = (" for run '{}' and channel number '{}'"
                        .format(self.run_id, self.channel_number))
        return check_instrument_specs(metadata_input, equipment, channel_info)

    def validate_e_property(self, param: tuple, metadata_input: float) -> bool:
        e_prop, min_range, max_range, units = param
        try:
            metadata_input = float(metadata_input)
        except (TypeError, ValueError):
            logger.error("The {} for run '{}' and channel number '{}' should "
                         "be a float. {}".format(e_prop, self.run_id,
                                                 self.channel_number, MSG_E_CHA))
            return False
        else:
            if metadata_input < min_range:
                logger.error("The {} for run '{}' and for channel number "
                             "'{}' should be positive."
                             .format(e_prop, self.run_id, self.channel_number))
                return False
            if metadata_input > max_range:
                msg = ''
                if 'resistance' in e_prop or 'voltage' in e_prop:
                    msg = ("A {} > {}{} is indicative of poor contact."
                           .format(e_prop, max_range, units))
                logger.warning("The {} for run '{}' and for channel number "
                               "'{}' should be less than {}{}. {}"
                               .format(e_prop, self.run_id, self.channel_number,
                                       max_range, units, msg))
            return True

    def validate_contact_resistance_end(self, metadata_input: float) -> bool:
        if metadata_input is not None:
            param = ("contact resistance (end)", 0, 3000, 'Ω')
            return self.validate_e_property(param, metadata_input)
        return True

    def validate_contact_resistance_start(self, metadata_input: float) -> bool:
        if metadata_input is not None:
            param = ("contact resistance (start)", 0, 3000, 'Ω')
            return self.validate_e_property(param, metadata_input)
        return True

    def validate_dc_end(self, metadata_input: float) -> bool:
        if metadata_input is not None:
            param = ("DC voltage (end)", 0, 0.1, 'V')
            return self.validate_e_property(param, metadata_input)
        return True

    def validate_dc_start(self, metadata_input: float) -> bool:
        if metadata_input is not None:
            param = ("DC voltage (start)", 0, 0.1, 'V')
            return self.validate_e_property(param, metadata_input)
        return True

    def validate_dipole_length(self, metadata_input: float) -> bool:
        param = ("dipole length", 0, 99999, 'm')
        return self.validate_e_property(param, metadata_input)

    def validate_negative_electrode_serial_number(self, metadata_input: str) -> bool:
        equipment = ("negative electrode (run '{}' and channel number '{}')"
                     .format(self.run_id, self.channel_number))
        return check_serial_number(metadata_input, equipment)

    def validate_positive_electrode_serial_number(self, metadata_input: str) -> bool:
        equipment = ("positive electrode (run '{}' and channel number: '{}')"
                     .format(self.run_id, self.channel_number))
        return check_serial_number(metadata_input, equipment)
